#
/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *				   D R D C O N F . C
 *
 *				Douglas P. Kingston III
 *
 *				Part of the MDQS Daemon
 *
 *				   Revision History
 *				   ----------------
 *	30 Mar 83  DPK	Added code to clear dv_failed.  Changed code
 *			that dchains structures.
 */

#include <stdio.h>
#include "queue.h"

extern	struct queue *quehead;
extern	struct device *devhead;
extern	char *Qconffile;
extern	char *MDQSuser;
extern	char *Sysmgr;
extern	char *Logfile;
extern	char *Errorfile;
extern	int Scanwait;
extern	int Procwait;
extern	int Openwait;
extern	int Delaytime;
extern	int Maxfailures;
extern	int Debug;

struct device *getdev();
struct device *zapdev();
struct queue *getque();
struct queue *zapque();
char	*xmalloc();
char	*strdup();
char	*stradd();
char	*ctime();
int	setconsole();

char	*console = NULL;


/*
 *			R D C O N F
 *
 *	Read and configure the daemon according to the configuration file
 */
rdconf ()
{
	struct device *dp;
	register struct queue *qp;
	register struct qmap *mp;
	struct qmap *nextp;		/* Used for linked list reversal */
	struct qmap *prevp;		/* Used for linked list reversal */
	char	linebuf[LINESIZ];
	int	filestate = 0;		/* 0=devs, 1=queues, 2=map */
	int	argcnt;
	char	**fieldp;
	FILE	*cf;

	/*****
	 ***** there should be some kind of protection against
	 ***** against simultaneous access of the configuration file...
	 *****/
	if(( cf = fopen (Qconffile, "r")) == NULL ) {
		complain ("Cannot open configuration file (%s: %s)",
			Qconffile, sys_errlist[errno]);
		return;
	}
	fseek (cf, (long)(-4), 2);
	fread (linebuf, 4, 1, cf);
	if (strncmp (linebuf, "EOF\n", 4) != 0
	  && strncmp (linebuf+1, "EOF", 3) != 0) {
	  	complain ("Configuration file does not end in 'EOF'");
	  	return;
	}
	fseek (cf, 0L, 0);

	/*  Mark all devices so we can tell who's been deleted */
	if( devhead != NULL )
		for (dp = devhead; dp != NULL; dp = dp->dv_nextdev) {
			dp->dv_status |= D_TOUCH;

			/*
			 *  Trash the old queue-->device mapping table
			 */
			for (mp=dp->dv_maphd; mp!=NULL; mp=mp->m_nextmap) {
				free (mp->m_server);
				free (mp);
			}
			dp->dv_maphd = NULL;
		}

	/*  Mark all queues so we can tell who's been deleted */
	if( quehead != NULL )
		for (qp = quehead; qp != NULL; qp = qp->q_nextq)
			qp->q_status |= Q_TOUCH;

	while (fgets (linebuf, LINESIZ, cf) != NULL) {

		argcnt = parse (linebuf, &fieldp);

		/*  "Blank" line or bad line, probably a comment line */
		if (argcnt <= 0)
			continue;

		/*  Lines that start with a "-" indicate end of a section */
		if (**fieldp == '-') {
			filestate++;
			continue;
		}

		if( filestate == 0 ) {
			/*
			 *  Process system parameters
			 */
			if( argcnt != 2 ) {
				complain( "Bad parameter specification '%s'", *fieldp );
				continue;
			}

			setparam( fieldp );
		} else if( filestate == 1 ) {
			/* Process a device definition */
			if (argcnt < 2) {
				complain ("Bad device definition %s", *fieldp);
				continue;
			}

			dp = getdev (*fieldp);
			if (dp == NULL) {
				/*
				 * We have a NEW device, get a fresh struct
				 */
				dp = (struct device *)xmalloc (sizeof (struct device));
				dp->dv_dname = strdup (*fieldp++);
				dp->dv_device = strdup (*fieldp++);
				dp->dv_status = dstatus( fieldp );
				dp->dv_actque = NULL;
				dp->dv_actreq = NULL;
				dp->dv_maphd = NULL;
				dp->dv_pid = 0;

				devsinit( dp );

				/* Insert new device at head of devices */
				dp->dv_nextdev = devhead;
				devhead = dp;
debug ("NEW Device (%s, %s, %s, %d)", dp->dv_dname, dp->dv_device, dp->dv_forms, dp->dv_status);
			} else {
				/*
				 * Hard case, check an already existing device
				 */
				fieldp++;
				if (strcmp (*fieldp, dp->dv_device) != SAME) {
					dp->dv_status |= D_RESET;
					free (dp->dv_device);
					dp->dv_device = strdup (*fieldp);
				}
				fieldp++;
				upddev ( dp, dstatus(fieldp));
				dp->dv_status &= ~D_TOUCH;
debug ("OLD Device (%s, %s, %s, %d)", dp->dv_dname, dp->dv_device, dp->dv_forms, dp->dv_status);
			}
		} else if (filestate == 2) {
			/*
			 *  Process a QUEUE definition
			 */
			if (argcnt == 0) {
				complain ("Bad queue definition %s", *fieldp);
				continue;
			}

			qp = getque (*fieldp);

			if (qp == NULL) {
				/*
				 *  This is a new QUEUE definiton
				 */
				qp = (struct queue *)xmalloc (sizeof (struct queue));
				qp->q_name = strdup (*fieldp++);
				qp->q_status = qstatus( fieldp );
				qp->q_readyreq = NULL;
debug ("NEW Queue (%s, %d)", qp->q_name, qp->q_status);

				/* Force qlists() to create empty qlistfile */
				qp->q_status |= Q_UPDATE;

				/* Insert new queue at head of queues */
				qp->q_nextq = quehead;
				quehead = qp;
			} else {
				/*
				 *  Handle an already existing queue
				 */
				fieldp++;
				updque ( qp, qstatus( fieldp ));
				qp->q_status &= ~Q_TOUCH;
			}
		} else if (filestate == 3) {
			if( argcnt < 3 ) {
				complain("Bad map definition %s", *fieldp);
				continue;
			}

			/*
			 *  Get the queue--->device mapping data
			 */
			if ((qp = getque (*fieldp)) == NULL ||
			    (qp->q_status & Q_TOUCH) ||		/* Lame duck */
			    (dp = getdev (*++fieldp)) == NULL ||
			    (dp->dv_status & D_TOUCH) )	{	/* Lame duck */
				complain( "Bad mapping (%s)", *fieldp );
				continue;
			}
			mp = (struct qmap *)xmalloc (sizeof (struct qmap));
			mp->m_q = qp;
			mp->m_server = stradd(++fieldp, argcnt - 2);
			mp->m_nextmap = dp->dv_maphd;	/* Insert at head */
			dp->dv_maphd = mp;
debug ("Got the mapping (%s, %s, %s)", qp->q_name, dp->dv_dname, mp->m_server);
		}
	}

	fclose( cf );

	/*
	 *  Zap deleted devices!
	 */
	if( devhead == NULL ) {
		complain( "No devices specified" );
		goto nodevs;
	}
	while( devhead->dv_status & D_TOUCH ) {
		devhead = zapdev( devhead );
		if (devhead == NULL) {
			complain( "All devices deleted" );
			goto nodevs;
		}
	}
	dp = devhead;
	dp->dv_failed = 0;
	while( dp->dv_nextdev != NULL ) {
		if( dp->dv_nextdev->dv_status & D_TOUCH )
			/** PITCH IT **/
			dp->dv_nextdev = zapdev( dp->dv_nextdev );
		else {
			/** KEEP IT **/
			dp = dp->dv_nextdev;
			dp->dv_failed = 0;
		}
	}
debug( "Devices zapped!" );

nodevs:
	/*
	 *  Zap deleted queues!
	 *	Warning:  This means zapping all the requests on that queue.
	 */
	if( quehead == NULL )
		bomb( "No queues specified" );
	while( quehead->q_status & Q_TOUCH ) {
		quehead = zapque( quehead );
		if( quehead == NULL )
			bomb( "All queues deleted" );
	}
	qp = quehead;
	while( qp->q_nextq != NULL ) {
		if( qp->q_nextq->q_status & Q_TOUCH )
			qp->q_nextq = zapque( qp->q_nextq );
		else
			qp = qp->q_nextq;
	}
debug( "Queues zapped!");

	/*
	 *  Reverse the chain of qmap structures on each device
	 */
	for (dp = devhead; dp != NULL; dp = dp->dv_nextdev) {
		mp = NULL;
		nextp = dp->dv_maphd;
		while (nextp != NULL) {
			prevp = mp;		/* previous gets current */
			mp = nextp;		/* current gets next */
			nextp = nextp->m_nextmap;  /* next gets next one */
			mp->m_nextmap = prevp;	/* burn our bridges... */
		}
		dp->dv_maphd = mp;
	}
debug( "----- System configured -----" );
}

/*
 *	S T R A D D ( )
 *
 *	Takes a pointer to an element in an array of string pointers
 *	and a count and returns a pointer to a string that is made
 *	by concatenating count strings with spaces.
 */
char  *
stradd  (strpp,  cnt)
char  	**strpp;
int	cnt;
{
	register      char	*fcp;
	register      char	*tcp;
	char	buf[LINESIZ];

	tcp = buf;
	while (cnt-- > 1) {
		fcp = *strpp++;
		while (*fcp)
			*tcp++ = *fcp++;
		*tcp++ = ' ';
	}
	fcp = *strpp;
	while (*fcp)
		*tcp++ = *fcp++;
	*tcp = 0;
	return (strdup(buf));
}

struct paramtab paramtab[] = {
	"scanwait",	INTVAR,		(char **) &Scanwait,	NULL,
	"procwait",	INTVAR,		(char **) &Procwait,	NULL,
	"openwait",	INTVAR,		(char **) &Openwait,	NULL,
	"delaytime",	INTVAR,		(char **) &Delaytime,	NULL,
	"debug",	INTVAR,		(char **) &Debug,	NULL,
	"maxfailures",	INTVAR,		(char **) &Maxfailures,	NULL,
	"console",	STRPVAR,	NULL,			setconsole,
	"sysmgr",	STRPVAR,	&Sysmgr,		NULL,
	"mdqsid",	STRPVAR,	&MDQSuser,		NULL,
	"logfile",	STRPVAR,	&Logfile,		NULL,
	"errorfile",	STRPVAR,	&Errorfile,		NULL,
	0,		0,		0,			0
};

setparam( fp )
char	**fp;
{
	struct paramtab *pp;

	for( pp = paramtab; pp->paramstr != 0; pp++ )
		if( strcmp( *fp, pp->paramstr ) == SAME ) {
			switch( pp->vartype ) {
			case INTVAR:
				if( pp->setfunc != NULL )
					(*pp->setfunc)( fp );
				if( pp->parmp != NULL )
					*((int *)pp->parmp) = atoi( *++fp );
				break;

			case STRPVAR:
				if( pp->setfunc != NULL )
					(*pp->setfunc)( fp );
				if( pp->parmp != NULL ) {
					if( *pp->parmp )
						free( *pp->parmp );
					*pp->parmp = strdup( *++fp );
				}
				break;
			}
			return;
		}
}

struct queue *
getque ( qname )
register char	*qname;
{
	register struct queue *qp;

	if (quehead == NULL)
		return ( NULL );
	for (qp = quehead; qp != NULL; qp = qp->q_nextq)
		if (strncmp ( qname, qp->q_name, QUESIZ) == SAME)
			return (qp);
	return ( NULL );
}

struct device *
getdev ( dname )
register char	*dname;
{
	register struct device *dp;

	if (devhead == NULL)
		return ( NULL );
	for (dp = devhead; dp != NULL; dp = dp->dv_nextdev)
		if (strcmp ( dname, dp->dv_dname) == SAME)
			return (dp);
	 return ( NULL );
}

dstatus( strpp )
register char	**strpp;
{
	register int	status;

	status = 0;
	while (*strpp != NULL) {
		if (strcmp (*strpp, "anyform") == SAME)
			status |= D_ANYFORM;
		else if (strcmp (*strpp, "skipmsg") == SAME)
			status |= D_SKIPMSG;
		else if (strcmp (*strpp, "roundrobin") == SAME)
			status |= D_RROBIN;
		else
			complain ("unknown device flag '%s'", *strpp);
		strpp++;
	}
	return (status);	
}

qstatus( strpp )
register char	**strpp;
{
	register int	status;

	status = 0;
	while (*strpp != NULL) {
		if(1)
			complain ("unknown queue flag '%s'", *strpp);
		strpp++;
	}
	return (status);
}

upddev( dp, status )
register struct device *dp;
int	status;
{
	if( dp->dv_status & D_RESET ) {
		/*  Clear and restart the device  */
		/*  Kill the current job */
		/**** UNDONE, dummy for now ****/
		dp->dv_status &= ~D_RESET;
	}
	dp->dv_status = (dp->dv_status & (D_RESET|D_TOUCH))
			| (status & ~(D_RESET|D_TOUCH));
}

updque( qp, status )
register struct queue *qp;
int	status;
{
	/**** UNDONE, dummy for now ****/
}

/*
 *	SETCONSOLE
 *  This function will check the validity of the filename
 *  supplied as an argument to the console parameter, and
 *  then open it as the console for the MDQS program.
 *  The fp argument will be pointing at the pointer to the
 *  parameter name.
 */
setconsole( fp )
register char	**fp;
{
	register int	fd;
	time_t	timenow;

	fp++;
	if (*fp == NULL)
		return;
	if (console != NULL && strcmp (*fp, console) == SAME)
		return;
	if( access( *fp, WACCESS ) < 0 ) {
		complain( "Cannot access %s for use as console (%s)",
			*fp, sys_errlist[errno] );
		return;
	}
	if(( fd = open( *fp, WRITE )) < 0 ) {
		complain( "Error opening new console %s (%s)",
			*fp, sys_errlist[errno] );
		return;
	}
	lseek (fd, 0L, 2);	/* Go to the end of the file */
	if (console != NULL)
		free (console);
	console = strdup (*fp);

	/*
	 *  We have a new console open, mention the switch on the
	 *  old console and then close it.
	 */
	time( &timenow );
	complain( "Switching console to %s at %s", *fp, ctime( &timenow ));
	fflush( stderr );
	close( 2 );
	dup2( fd, 2 );
	dup2( 2, 1 );
	close( fd );
	complain( "Console opened at %s", ctime( &timenow ));
}
