#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		    Imagen Laser Printer Server (imserver)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "queue.h"
#include "qreplys.h"

#if BSD42 | BSD41c | V7 | V6
#include <sgtty.h>
struct sgttyb ttyb;
#include <sys/file.h>
#endif
#if BRLUNIX | DPWUNIX
#include <sgtty.h>
struct sg_brl ttyb;
#endif

/* pipe() result indexes */
#define	PREAD	0
#define	PWRITE	1

#define NIPSARGS	50
char	*ips	= "/usr/local/lib/imagen/ips";
char	*ipsargs[NIPSARGS] = {"imagen", "-s", 0 };
char	ipsargc = 2;
char	*language;
struct	request r;
int	acctfd;
int	pid;
int	lpoptions;
int	speed = B9600;
int	(*process)();
char	*workdir;
char	*title;
char	*user;

char	inputfile[PATHSIZ];		/* Name of datafile with cmds */

extern	char *Qdatadir;

extern	char	*index();
extern	char	*malloc();
extern	char	*strdup();
extern	struct	request *gethdr();
extern	char	*xmalloc();
extern	char	*fixstr();
extern	int	impress();
extern	int	tektronix();
extern	int	printer();
extern	int	daisy();
int	dorestart();
int	doflush();
int	badpipe();

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	linebuf[LINESIZ];
	char	*cp;
	int	fin;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	user = strdup ("UnknownUser");
	title = strdup ("UnknownTitle");

	arginit (argc, argv);
	setuid(getuid());		/* return to real uid */
	
	if (gethdr (stdin, &r) == (struct request *) NULL) {
		fprintf (stderr, "IMSERVER: Error reading request header\n");
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	/* For use later... */
	gtty (1, &ttyb);

	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'D':
			workdir = strdup(&linebuf[1]);
			break;

		case 'F':
			if (workdir == 0)
				workdir = "/tmp";
			if (chdir (workdir) < 0)
				chdir ("/tmp");
			fin = open( &linebuf[1], READ );
			if( fin < 0 ) {
				fprintf( stderr, "IMSERVER: Can't open input file '%s'\n", &linebuf[1] );
				break;
			} else {
				sendit(fin);
				close (fin);
			}
			if (lpoptions & NOKEEP)
				unlink (&linebuf[1]);
			break;

		case 'I':
			sprintf( inputfile, "%s/%s", Qdatadir, &linebuf[1] );
			fin = open( inputfile, READ );
			if( fin < 0 ) {
				fprintf( stderr, "IMSERVER: Can't open input file '%s'\n", inputfile );
				retmsg (RP_FATAL, "Cannot open input file");
				exit( 99 );
			} else {
				sendit(fin);
				close (fin);
			}
			unlink( inputfile );
			break;

		case 'O':
			lpoptions = atoi(&linebuf[0]);
			break;

		case 'T':
			title = strdup(&linebuf[1]);
			break;

		case 'X':
			setargs( &linebuf[1] );
			break;

		case 'U':
			user = strdup(&linebuf[1]);
			break;

		default:
			/* Ignore unknown control lines */
			break;
		}
	}
	retmsg (RP_OK, 0);
	exit (0);
}

/*
 *  S E N D I T (  )
 *
 *  Send an input file to the imagen.
 */
sendit(fin)
register int fin;
{
	struct	status status;
	int	pfds[2];
	int	i;

	/* Fix up the imagen comm line */
	ioinit(1);

	/*
	 *	Now we are ready to start the imagen transfer process.
	 *	We are already running with our UID/GID
	 *	the same as the invoker.
	 */
	if( pipe(pfds) < 0 ) {
		retmsg (RP_RESTART, "pipe() failed");
		exit(0);
	}

	if(( pid = fork()) == -1 ) {
		retmsg (RP_RESTART, "Fork() failed");
		exit (0);
	}
	else if( pid == 0 ) {
		/*
		 *  Child
		 */
		close (pfds[PWRITE]);
		dup2 (pfds[PREAD], 0);	/* Stdin from pipe */
		close (pfds[PREAD]);

		/* stdin is imserver, stdout is imagen, stderr is console */
#ifndef _NFILE
		for( i = getdtablesize()-1; i > 2; i--)
#else
		for( i = _NFILE-1; i > 2; i--)
#endif
			close( i );

		if (workdir == 0)
			workdir = "/tmp";
		if( chdir( workdir ) < 0 )
			chdir( "/tmp" );	/* Its a safe default */

		execv( ips, ipsargs );
fprintf( stderr, "IMSERVER: exec of %s failed (%s)\n",
			ips, sys_errlist[errno]);
		abort();
	} else {
		int	pagecnt;
		int	retstat;
		FILE	*ifp, *ofp;

		/*
		 *  Parent
		 */
		signal (SIGPIPE, badpipe);
		close (pfds[PREAD]);
		ofp = fdopen (pfds[PWRITE], "w");
		ifp = fdopen (fin, "r");
		doheader (ofp);
		pagecnt = (*process)(ifp, ofp);
		fclose (ofp);

		wait( &retstat );
		status.ret = (retstat>>8) & 0xff;
		status.pstat = retstat & 0xff;
		if( status.pstat ) {
			retmsg (RP_HOLD, "Imagen (ips) process bombed");
			fprintf( stderr, "IMSERVER: request %s, ips bombed\n",
				 r.r_name );
			account(pagecnt, language, &status, user);
			exit (0);
		}
		account(pagecnt, language, &status, user);
	}
}

doheader (fout)
FILE	*fout;
{
	register int c;
	char	*cp;
	time_t timenow;

	time (&timenow);
	fputs ("@Document(", fout);
	cp = fixstr(user);
	fprintf (fout, "Owner %s", cp);
	free (cp);
	cp = fixstr(title);
	fprintf (fout, ", Name %s", cp);
	free (cp);
	cp = fixstr(ctime(&r.r_tsubmit));
	fprintf (fout, ", Submitted %s", cp);
	free (cp);
	cp = fixstr(ctime (&r.r_tstart));
	fprintf (fout, ", Started %s", cp);
	free (cp);
	cp = fixstr(ctime (&timenow));
	fprintf (fout, ", Printed %s", cp);
	free (cp);
	fprintf (fout, ", Language %s", language);
	if (strcmp(language, "impress"))
		fputs (", Pagereversal", fout);
	fputs (", JobHeader, JamResistance)", fout);

	fflush(fout);
	if (ferror (fout)) {
		retmsg (RP_FATAL, "Error writing to imagen");
		exit (0);
	}
}

/*
 *		S E T A R G S ( )
 *	This function is responsible for building the argument
 *	variable vector.  Pass it a string to be installed.
 */
setargs( varstr )
register char	*varstr;
{
	if( ipsargc < NIPSARGS - 1 ) {
		ipsargs[ipsargc] = strdup(varstr);
		ipsargs[++ipsargc] = 0;
	}
}

/*
 *		X M A L L O C ( )
 *	Just malloc(), but it always returns a valid pointer or exits
 */
char *
xmalloc( size )
int	size;
{
	register char	*cp;

	if(( cp = malloc( size )) == NULL ) {
		fprintf( stderr, "IMSERVER: No free memory!\n");
		retmsg(RP_FATAL, "No free memory");
		if( *inputfile )
			unlink( inputfile );
		exit( 0 );
	}
	return( cp );
}

/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 */
dorestart()
{
	retmsg (RP_RESTART, "RESTART requested");
	if( pid )
		kill( pid, SIGKILL );
	exit( 0 );
}


/*
 *	BADPIPE ()
 *
 *  This function is called when the pipe to the child breaks
 */
badpipe()
{
	retmsg (RP_RESTART, "Pipe to IPS broke");
	if( pid > 1)
		kill( pid, SIGKILL );
	exit( 0 );
}


/*
 *	DOFLUSH ()
 *
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	retmsg (RP_FLUSH, "Job flushed on request");
	if( pid );
		kill( pid, SIGKILL );
	exit( 0 );
}

/*
 *	A R G I N I T
 *
 *	Parse args of lineprinter like servers
 */
arginit(argc, argv)
int	argc;
char	**argv;
{
	int	baud;

	if(argc < 3) {
		retmsg (RP_FATAL, "Imserver called improperly");
		err ("Missing language argument");
		exit (1);
	}
	if (strcmp (argv[1], "impress") == 0)
		process = impress;
	else if (strcmp (argv[1], "tektronix") == 0)
		process = tektronix;
	else if (strcmp (argv[1], "printer") == 0)
		process = printer;
	else if (strcmp (argv[1], "daisy") == 0)
		process = daisy;
	else {
		retmsg (RP_FATAL, "Imserver called with bad language");
		err ("Bad language argument");
		exit (1);
	}
	language = argv[1];


#if BSD42 | SYS3
	if ((acctfd = open (argv[2], 1, O_APPEND)) < 0) {
#else
	if ((acctfd = open (argv[2], 2)) < 0) {
#endif
		retmsg (RP_FATAL, "Imserver cannot open accounting file");
		err ("Cannot open accounting file");
		exit (1);
	}
	if(argc > 3) {
		if ((baud = atoi (argv[4])) == 0) {
			err ("Bad argument '%s'", argv[4]);
			return;
		}
		switch (baud) {
		case 134:   speed = B134; break;
		case 150:   speed = B150; break;
		case 300:   speed = B300; break;
		case 600:   speed = B600; break;
		case 1200:  speed = B1200; break;
		case 1800:  speed = B1800; break;
		case 2400:  speed = B2400; break;
		case 4800:  speed = B4800; break;
		case 9600:  speed = B9600; break;
		}
	}
}

/*
 *  I O I N I T  (  )
 *
 *  Stty the line appropriately for the Imagen
 */
ioinit (fd)
int	fd;
{
	ttyb.sg_ispeed = ttyb.sg_ospeed = speed;
	ttyb.sg_flags = RAW;
	stty (fd, &ttyb);
}

/*
 *	F I X S T R  (  )
 *
 *  Sanitize strings for Imagen document control
 */
char *
fixstr (cp)
register char *cp;
{
	char	tmpbuf[300];
	register char	*tp;

	tp = tmpbuf;
	*tp++ = '"';
	while (*cp) {
		switch (*cp) {
		case '"':		/* Swallow */
			cp++;
			break;
		case '@':
			*tp++ = ' ';
			*tp++ = 'a'; *tp++ = 't';
			*tp++ = ' ';
			cp++;
			break;
		default:
			*tp++ = *cp++;
			continue;

		}
	}
	*tp++ = '"';
	*tp = 0;
	return (strdup (tmpbuf));
}

account (pagecnt, language, status, user)
int	pagecnt;
char	*language;
struct status	*status;
char	*user;
{
	char	buf[512];
	int	len;
	time_t	timenow;

	time (&timenow);
	sprintf(buf, "%d %c %o/%o %12.12s %s\n", pagecnt, *language,
		status->pstat, status->ret, ctime(&timenow)+4, user);
	len = strlen(buf);
	lseek (acctfd, 0L, 2);
	if (write (acctfd, buf, len) != len) {
		retmsg (RP_FATAL, "Imserver write to accounting file");
		err ("Cannot write to accounting file");
		exit (1);
	}
}
