#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *			 Process Server  (PROCSERVER)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "queue.h"
#include "qreplys.h"

struct	request r;
char	*output = "/dev/null";
char	*title = 0;
char	*user = "UnknownUser";
char	*workdir = "/tmp";
char	datafile[PATHSIZ];
int	useshell = 0;			/* non-zero if you must "sh -c" */
int	pid;

/* 100 arguments ought to be enough ... */
#define NARGS	100
char	*envp[] = { 0 };
char	*args[NARGS] = { 0 };
int	argcnt = 0;

extern	char *Qdatadir;

char	*index();
char	*xmalloc();
char	*malloc();
char	*strdup();
int	dorestart();
int	doflush();
struct	request *gethdr();

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	linebuf[4*LINESIZ];
	char	*cp;
	int	fin;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	arginit(argc, argv);
	title = strdup ("UnknownTitle");
	
	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err("PROCSERVER: Error reading request");
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	statmsg ("running");
	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'D':
			workdir = strdup (&linebuf[1]);
			break;

		case 'F':
			if ((fin = open( &linebuf[1], READ )) < 0)
				printf( "PROCSERVER: Can't open command file '%s'\n", &linebuf[1]);
			else {
				process (fin);
				close (fin);
			}
			break;

		case 'I':
			sprintf( datafile, "%s/%s", Qdatadir, &linebuf[1] );
			if ((fin = open( datafile, READ )) < 0)
				printf( "PROCSERVER: Can't open command file '%s'\n", datafile );
			else {
				process(fin);
				close (fin);
			}
			break;

		case 'O':
			setoptions( &linebuf[1] );
			break;

		case 'P':
			output = strdup(&linebuf[1]);
			break;

		case 'T':
			if (title)
				free(title);
			title = strdup (&linebuf[1]);
			break;

		case 'U':
			user = strdup(&linebuf[1]);
			break;

		default:
			/* Ignore unknown control lines */
			break;
		}
	}
}

process (ifd)
int	ifd;
{
	int	status;
	char	**av;
	char	*tmpargs[NARGS];
	int	i = 0;

	av = args;
	while (*av) {
		if (strcmp (*av, "$T") == 0)
			tmpargs[i++] = title;
		else if (strcmp (*av, "$U") == 0)
			tmpargs[i++] = user;
		else
			tmpargs[i++] = *av;
		av++;
	}
	tmpargs[i] = 0;

	/*
	 *	Now we are read to start the process.  We are already
	 *	running with our UID/GID the same as the invoker.
	 */
	switch (pid = fork()) {
	case -1:
		retmsg (RP_RESTART, "Fork() failed");
		exit (0);

	case 0:
		/* Child */
		if (chdir (workdir) < 0) {
			err("PROCSERVER: Cannot chdir to '%s'", workdir);
			exit (98);
		}

		dup2( ifd, 0 );			/* Standard input is cmds */
		if(( i = open( output, WRITE )) < 0 ) {
			if( errno != ENOENT
			  || (i = creat( output, 0644 )) < 0) {
err("PROCSERVER: couldn't open output '%s', request refused", output );
				exit( 99 );
			}
		}
		lseek (i, 0L, 2);	/* Go to the end of the logfile */
		dup2( i, 1 );
		dup2( 1, 2 );
#ifndef _NFILE
		for( i = getdtablesize()-1; i > 2; i--)
#else
		for( i = _NFILE-1; i > 2; i--)
#endif
			close( i );

		if (useshell) {
			char	buf[PATHSIZ*2];

			buf[0] = 0;
			for (av = tmpargs; *av; av++) {
				strcat (buf, *av);
				strcat (buf, " ");
			}
			execl ("/bin/sh", "proc-sh", "-c", buf, 0);
		} else
			execve( *args, tmpargs, envp );
err("PROCSERVER: exec of %s failed (%s)", *args, sys_errlist[errno]);
		abort();

	default:
		/* Parent */
		statmsg("waiting for child");
		wait( &status );
		unlink( datafile );
		if( status&0xff ) {
err("PROCSERVER: request %s, '%s' aborted", r.r_name, *args );
			retmsg (RP_FATAL, "process '%s' aborted", *args);
			exit (0);
		}
		if( (status>>8) & 0xff ) {
err("PROCSERVER: request %s, '%s' returned error", r.r_name, *args );
			retmsg (RP_FATAL, "process '%s' returned error", *args);
			exit (0);
		}
	}

	retmsg (RP_OK, 0);
	exit (0);
}

arginit(argc, argv)
int	argc;
char	**argv;
{
	register char	*arg;

	argv++;
	while (*argv != 0) {
		arg = *argv;
		if (*(arg++) != '-')
			break;
		argv++;
		switch (*arg) {
		case 's':
			useshell++;
			break;
		default:
			err ("PROCSERVER: unknown arg '-%s'", arg);
			break;
		}
	}
	while (*argv != 0) {
		args[argcnt++] = *(argv++);
		args[argcnt] = 0;
	}
}

/*
 *		S E T O P T I O N S  ( )
 */
setoptions (optstr)
char *optstr;
{
}

/*
 *		S E T E N V ( )
 *	This function is responsible for building the environment
 *	variable vector.  Pass it an environment string to be installed.
 */
setenv( envstr )
register char	*envstr;
{
	static	int	nused = 0;

	if( nused < (NARGS - 1)) {
		envp[nused++] = strdup (envstr);
		envp[nused] = 0;
	}
}

/*
 *		X M A L L O C ( )
 *	Just malloc(), but it always returns a valid pointer or exits
 */
char *
xmalloc( size )
int	size;
{
	register char	*cp;

	if(( cp = malloc( size )) == NULL ) {
		err("PROCSERVER: No free memory!");
		retmsg (RP_FATAL, "No free memory");
		if( *datafile )
			unlink( datafile );
		exit( 0 );
	}
	return( cp );
}

/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 */
dorestart()
{
	retmsg (RP_RESTART, "RESTART received");
		if( pid )
#if V6
			kill( pid, SIGKILL );
#endif
#if BSD42 | BSD41 | BSD41c | BRLUNIX
			killpg( pid, SIGKILL );
#endif
#if SYS3 | V7
			kill( 0, SIGKILL );
#endif
	exit( 0 );
}


/*
 *	DOFLUSH ()
 *
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	retmsg (RP_FLUSH, "Server received FLUSH request");
#if V6
	if( pid )
		kill( pid, SIGKILL );
#endif
#if BSD42 | BSD41c | BRLUNIX
	killpg( pid, SIGKILL );
#endif
#if SYS3 | BSD41 | V7
	kill( 0, SIGKILL );
#endif
	exit( 0 );
}
